//
//  MainComposeTextView.swift
//  Tusker
//
//  Created by Shadowfacts on 8/29/20.
//  Copyright © 2020 Shadowfacts. All rights reserved.
//

import SwiftUI
import Pachyderm

struct MainComposeTextView: View {
    @ObservedObject var draft: Draft
    let placeholder: Text
    
    let minHeight: CGFloat = 150
    @State private var height: CGFloat?
    @State private var becomeFirstResponder: Bool = false
    @State private var hasFirstAppeared = false
    
    var body: some View {
        ZStack(alignment: .topLeading) {
            Color(UIColor.secondarySystemBackground)
            
            if draft.text.isEmpty {
                placeholder
                    .font(.system(size: 20))
                    .foregroundColor(.secondary)
                    .offset(x: 4, y: 8)
            }
            
            MainComposeWrappedTextView(
                text: $draft.text,
                visibility: draft.visibility,
                becomeFirstResponder: $becomeFirstResponder
            ) { (textView) in
                self.height = max(textView.contentSize.height, minHeight)
            }
        }
        .frame(height: height ?? minHeight)
        .onAppear {
            if !hasFirstAppeared {
                hasFirstAppeared = true
                becomeFirstResponder = true
            }
        }
    }
}

struct MainComposeWrappedTextView: UIViewRepresentable {
    typealias UIViewType = UITextView
    
    @Binding var text: String
    let visibility: Status.Visibility
    @Binding var becomeFirstResponder: Bool
    var textDidChange: (UITextView) -> Void
    
    @EnvironmentObject var uiState: ComposeUIState
    @State var visibilityButton: UIBarButtonItem?
    
    func makeUIView(context: Context) -> UITextView {
        let textView = WrappedTextView()
        textView.delegate = context.coordinator
        textView.isEditable = true
        textView.backgroundColor = .clear
        textView.font = .systemFont(ofSize: 20)
        textView.textContainer.lineBreakMode = .byWordWrapping
        context.coordinator.textView = textView
        
        uiState.autocompleteHandler = context.coordinator
        
        let visibilityButton = UIBarButtonItem(image: UIImage(systemName: visibility.imageName), style: .plain, target: nil, action: nil)
        updateVisibilityMenu(visibilityButton)
        let toolbar = UIToolbar()
        toolbar.translatesAutoresizingMaskIntoConstraints = false
        toolbar.items = [
            UIBarButtonItem(title: "CW", style: .plain, target: nil, action: #selector(ComposeHostingController.cwButtonPressed)),
            visibilityButton,
            UIBarButtonItem(barButtonSystemItem: .flexibleSpace, target: nil, action: nil),
        ] + createFormattingButtons(coordinator: context.coordinator) + [
            UIBarButtonItem(barButtonSystemItem: .flexibleSpace, target: nil, action: nil),
            UIBarButtonItem(title: "Drafts", style: .plain, target: nil, action: #selector(ComposeHostingController.draftsButtonPresed)),
        ]
        textView.inputAccessoryView = toolbar
        // can't modify @State during view update
        DispatchQueue.main.async {
            self.visibilityButton = visibilityButton
        }
        
        NotificationCenter.default.addObserver(context.coordinator, selector: #selector(Coordinator.keyboardWillShow(_:)), name: UIResponder.keyboardWillShowNotification, object: nil)
        NotificationCenter.default.addObserver(context.coordinator, selector: #selector(Coordinator.keyboardWillHide(_:)), name: UIResponder.keyboardWillHideNotification, object: nil)
        NotificationCenter.default.addObserver(context.coordinator, selector: #selector(Coordinator.keyboardDidHide(_:)), name: UIResponder.keyboardDidHideNotification, object: nil)
        
        return textView
    }
    
    private func createFormattingButtons(coordinator: Coordinator) -> [UIBarButtonItem] {
        guard Preferences.shared.statusContentType != .plain else {
            return []
        }
        
        var formatButtons = StatusFormat.allCases.map { (format) -> UIBarButtonItem in
            let item: UIBarButtonItem
            if let image = format.image {
                item = UIBarButtonItem(image: image, style: .plain, target: coordinator, action: #selector(Coordinator.formatButtonPressed(_:)))
            } else if let (str, attributes) = format.title {
                item = UIBarButtonItem(title: str, style: .plain, target: coordinator, action: #selector(Coordinator.formatButtonPressed(_:)))
                item.setTitleTextAttributes(attributes, for: .normal)
                item.setTitleTextAttributes(attributes, for: .highlighted)
            } else {
                fatalError("StatusFormat must have either an image or a title")
            }
            item.tag = StatusFormat.allCases.firstIndex(of: format)!
            item.accessibilityLabel = format.accessibilityLabel
            return item
        }
        
        for i in (1..<StatusFormat.allCases.count).reversed() {
            let spacer = UIBarButtonItem(barButtonSystemItem: .fixedSpace, target: nil, action: nil)
            spacer.width = 8
            formatButtons.insert(spacer, at: i)
        }
        
        return formatButtons
    }
    
    private func updateVisibilityMenu(_ visibilityButton: UIBarButtonItem) {
        let elements = Status.Visibility.allCases.map { (visibility) -> UIMenuElement in
            let state = visibility == self.visibility ? UIMenuElement.State.on : .off
            return UIAction(title: visibility.displayName, image: UIImage(systemName: visibility.unfilledImageName), identifier: nil, discoverabilityTitle: nil, attributes: [], state: state) { (_) in
                self.uiState.draft.visibility = visibility
            }
        }
        visibilityButton.menu = UIMenu(title: "", image: nil, identifier: nil, options: [], children: elements)
    }
    
    func updateUIView(_ uiView: UITextView, context: Context) {
        if context.coordinator.skipSettingTextOnNextUpdate {
            context.coordinator.skipSettingTextOnNextUpdate = false
        } else {
            uiView.text = text
        }
        
        if let visibilityButton = visibilityButton {
            visibilityButton.image = UIImage(systemName: visibility.imageName)
            updateVisibilityMenu(visibilityButton)
        }
        context.coordinator.text = $text
        context.coordinator.didChange = textDidChange
        context.coordinator.uiState = uiState
        
        // wait until the next runloop iteration so that SwiftUI view updates have finished and
        // the text view knows its new content size
        DispatchQueue.main.async {
            self.textDidChange(uiView)
            
            if becomeFirstResponder {
                // calling becomeFirstResponder during the SwiftUI update causes a crash on iOS 13
                uiView.becomeFirstResponder()
                // can't update @State vars during the SwiftUI update
                becomeFirstResponder = false
            }
        }
    }
    
    func makeCoordinator() -> Coordinator {
        return Coordinator(text: $text, uiState: uiState, didChange: textDidChange)
    }
    
    class WrappedTextView: UITextView {
        private let formattingActions = [#selector(toggleBoldface(_:)), #selector(toggleItalics(_:))]
        
        override func canPerformAction(_ action: Selector, withSender sender: Any?) -> Bool {
            if formattingActions.contains(action) {
                return Preferences.shared.statusContentType != .plain
            }
            
            return super.canPerformAction(action, withSender: sender)
        }
        
        override func toggleBoldface(_ sender: Any?) {
            (delegate as! Coordinator).applyFormat(.bold)
        }
        
        override func toggleItalics(_ sender: Any?) {
            (delegate as! Coordinator).applyFormat(.italics)
        }
        
        override func validate(_ command: UICommand) {
            super.validate(command)
            
            if formattingActions.contains(command.action),
               Preferences.shared.statusContentType != .plain {
                command.attributes.remove(.disabled)
            }
        }
    }
    
    class Coordinator: NSObject, UITextViewDelegate, ComposeAutocompleteHandler, ComposeTextViewCaretScrolling {
        weak var textView: UITextView?
        var text: Binding<String>
        var didChange: (UITextView) -> Void
        var uiState: ComposeUIState
        var caretScrollPositionAnimator: UIViewPropertyAnimator?
        
        var skipSettingTextOnNextUpdate = false
        
        init(text: Binding<String>, uiState: ComposeUIState, didChange: @escaping (UITextView) -> Void) {
            self.text = text
            self.didChange = didChange
            self.uiState = uiState
        }
        
        func textViewDidChange(_ textView: UITextView) {
            text.wrappedValue = textView.text
            didChange(textView)
            
            ensureCursorVisible(textView: textView)
        }
        
        @objc func formatButtonPressed(_ sender: UIBarButtonItem) {
            let format = StatusFormat.allCases[sender.tag]
            applyFormat(format)
        }
        
        func applyFormat(_ format: StatusFormat) {
            guard let textView = textView,
                  textView.isFirstResponder,
                  let insertionResult = format.insertionResult else {
                return
            }
            
            let currentSelectedRange = textView.selectedRange
            if currentSelectedRange.length == 0 {
                textView.insertText(insertionResult.prefix + insertionResult.suffix)
                textView.selectedRange = NSRange(location: currentSelectedRange.location + insertionResult.insertionPoint, length: 0)
            } else {
                let start = textView.text.index(textView.text.startIndex, offsetBy: currentSelectedRange.lowerBound)
                let end = textView.text.index(textView.text.startIndex, offsetBy: currentSelectedRange.upperBound)
                let selectedText = textView.text[start..<end]
                textView.insertText(String(insertionResult.prefix + selectedText + insertionResult.suffix))
                textView.selectedRange = NSRange(location: currentSelectedRange.location + insertionResult.prefix.utf16.count, length: currentSelectedRange.length)
            }
        }
        
        @objc func keyboardWillShow(_ notification: Foundation.Notification) {
            uiState.delegate?.keyboardWillShow(accessoryView: textView!.inputAccessoryView!, notification: notification)
        }
        
        @objc func keyboardWillHide(_ notification: Foundation.Notification) {
            uiState.delegate?.keyboardWillHide(accessoryView: textView!.inputAccessoryView!, notification: notification)
        }
        
        @objc func keyboardDidHide(_ notification: Foundation.Notification) {
            uiState.delegate?.keyboardDidHide(accessoryView: textView!.inputAccessoryView!, notification: notification)
        }
        
        func textViewDidBeginEditing(_ textView: UITextView) {
            uiState.autocompleteHandler = self
            updateAutocompleteState()
        }
        
        func textViewDidEndEditing(_ textView: UITextView) {
            updateAutocompleteState()
        }
        
        func textViewDidChangeSelection(_ textView: UITextView) {
            // Setting the text view's text causes it to move the cursor to the end (though only
            // when the text contains an emoji :/), so skip setting the text on the next SwiftUI update
            // that's triggered by setting the autocomplete state.
            skipSettingTextOnNextUpdate = true
            self.updateAutocompleteState()
        }
        
        func autocomplete(with string: String) {
            guard let textView = textView,
                  let text = textView.text,
                  let (lastWordStartIndex, _) = findAutocompleteLastWord() else {
                return
            }
            
            let distanceToEnd = text.utf16.count - textView.selectedRange.upperBound
            
            let characterBeforeCursorIndex = text.utf16.index(text.startIndex, offsetBy: textView.selectedRange.upperBound)
            
            let insertSpace: Bool
            if distanceToEnd > 0 {
                let charAfterCursor = text[characterBeforeCursorIndex]
                insertSpace = charAfterCursor != " " && charAfterCursor != "\n"
            } else {
                insertSpace = true
            }
            let string = insertSpace ? string + " " : string
            
            textView.text.replaceSubrange(lastWordStartIndex..<characterBeforeCursorIndex, with: string)
            self.textViewDidChange(textView)
            self.updateAutocompleteState()
            
            // keep the cursor at the same position in the text, immediately after what was inserted
            // if we inserted a space, move the cursor 1 farther so it's immediately after the pre-existing space
            let insertSpaceOffset = insertSpace ? 0 : 1
            textView.selectedRange = NSRange(location: textView.text.utf16.count - distanceToEnd + insertSpaceOffset, length: 0)
        }
        
        private func updateAutocompleteState() {
            guard let textView = textView,
                  let text = textView.text,
                  let (lastWordStartIndex, foundFirstAtSign) = findAutocompleteLastWord() else {
                uiState.autocompleteState = nil
                return
            }
            
            let triggerChars: [Character] = ["@", ":", "#"]
            if lastWordStartIndex > text.startIndex {
                // if the character before the "word" beginning is a valid part of a "word",
                // we aren't able to autocomplete
                let c = text[text.index(before: lastWordStartIndex)]
                if isPermittedForAutocomplete(c) || triggerChars.contains(c) {
                    uiState.autocompleteState = nil
                    return
                }
            }
            
            let characterBeforeCursorIndex = text.utf16.index(text.startIndex, offsetBy: textView.selectedRange.upperBound)
            
            if lastWordStartIndex >= text.startIndex {
                let lastWord = text[lastWordStartIndex..<characterBeforeCursorIndex]
                let exceptFirst = lastWord[lastWord.index(after: lastWord.startIndex)...]
                
                // periods are only allowed in mentions in the domain part
                if lastWord.contains(".") {
                    if lastWord.first == "@" && foundFirstAtSign {
                        uiState.autocompleteState = .mention(String(exceptFirst))
                    } else {
                        uiState.autocompleteState = nil
                    }
                    return
                }
                
                switch lastWord.first {
                case "@":
                    uiState.autocompleteState = .mention(String(exceptFirst))
                case ":":
                    uiState.autocompleteState = .emoji(String(exceptFirst))
                case "#":
                    uiState.autocompleteState = .hashtag(String(exceptFirst))
                default:
                    uiState.autocompleteState = nil
                }
            } else {
                uiState.autocompleteState = nil
            }
        }
        
        private func isPermittedForAutocomplete(_ c: Character) -> Bool {
            return (c >= "a" && c <= "z") || (c >= "A" && c <= "Z") || (c >= "0" && c <= "9") || c == "_"
        }
        
        private func findAutocompleteLastWord() -> (index: String.Index, foundFirstAtSign: Bool)? {
            guard let textView = textView,
                  textView.isFirstResponder,
                  textView.selectedRange.length == 0,
                  textView.selectedRange.upperBound > 0,
                  let text = textView.text,
                  text.count > 0 else {
                return nil
            }
            
            let characterBeforeCursorIndex = text.utf16.index(text.startIndex, offsetBy: textView.selectedRange.upperBound)
            
            var lastWordStartIndex = text.index(before: characterBeforeCursorIndex)
            var foundFirstAtSign = false
            while true {
                let c = text[lastWordStartIndex]
                
                if !isPermittedForAutocomplete(c) {
                    if foundFirstAtSign {
                        if c != "@" {
                            // move the index forward by 1, so that the first char of the substring is the 1st @ instead of whatever comes before it
                            lastWordStartIndex = text.index(after: lastWordStartIndex)
                        }
                        break
                    } else {
                        if c == "@" {
                            foundFirstAtSign = true
                        } else if c != "." {
                            // periods are allowed for domain names in mentions
                            break
                        }
                    }
                }
                
                if lastWordStartIndex > text.startIndex {
                    lastWordStartIndex = text.index(before: lastWordStartIndex)
                } else {
                    break
                }
            }
            
            return (lastWordStartIndex, foundFirstAtSign)
        }
    }
}
